import json
import os
import logging

import pymongo
import redis
import xappy


log = logging.getLogger("charm.index")


def setup_indexer(indexer):
    indexer.add_field_action(
        'name', xappy.FieldActions.INDEX_FREETEXT, weight=10,
        language='en')
    indexer.add_field_action('name', xappy.FieldActions.STORE_CONTENT)
    indexer.add_field_action('short_url', xappy.FieldActions.STORE_CONTENT)
    indexer.add_field_action('label', xappy.FieldActions.STORE_CONTENT)

    indexer.add_field_action(
        'summary', xappy.FieldActions.INDEX_FREETEXT, weight=5,
        language='en')
    indexer.add_field_action('summary', xappy.FieldActions.STORE_CONTENT)

    indexer.add_field_action(
        'owner', xappy.FieldActions.INDEX_FREETEXT, weight=5,
        language='en')

    indexer.add_field_action(
        'description', xappy.FieldActions.INDEX_FREETEXT, weight=3,
        language='en')

    indexer.add_field_action(
        'config', xappy.FieldActions.INDEX_FREETEXT, language='en')
    indexer.add_field_action(
        'relations', xappy.FieldActions.INDEX_FREETEXT, language='en')
    indexer.add_field_action(
        'changes', xappy.FieldActions.INDEX_FREETEXT, language='en')


def index_charm(indexer, charm):
    doc = xappy.UnprocessedDocument()
    doc.fields.append(xappy.Field("name", charm["name"]))
    doc.fields.append(xappy.Field("summary", charm["summary"]))
    doc.fields.append(xappy.Field("description", charm["description"]))
    doc.fields.append(xappy.Field("owner", charm["owner"]))
    doc.fields.append(xappy.Field("short_url", charm["short_url"]))
    doc.fields.append(xappy.Field("label", charm["label"]))

    if "config" in charm \
           and 'options' in charm['config'] \
           and charm['config']['options']:
        config_text = []

        for key, option in charm["config"]["options"].items():
            config_text.append(key)
            config_text.append(option.get("description", ""))
            #config_text.append(option.get("default"))
        doc.fields.append(xappy.Field("config", " ".join(config_text)))

    relation_text = []
    if "requires" in charm and charm["requires"]:
        for key, option in charm["requires"].items():
            relation_text.append(key)
            relation_text.append(option["interface"])
    if "provides" in charm and charm["provides"]:
        for key, option in charm["provides"].items():
            if not isinstance(option, dict):
                log.warning("invalid charm provides %s", charm['branch_spec'])
                continue
            relation_text.append(key)
            relation_text.append(option["interface"])
    if relation_text:
        doc.fields.append(xappy.Field("relations", " ".join(relation_text)))

    change_text = []
    for change in charm.get("changes", ()):
        change_text.append(change['message'])
        change_text.append(change['committer'])
    if change_text:
        doc.fields.append(xappy.Field("changes", " ".join(change_text)))

    doc.id = charm["_id"]
    indexer.replace(doc)
    indexer.flush()


def index_repo(db, indexer, root_dir):
    charms = os.listdir(root_dir)

    for c in charms:
        charm_dir = os.path.join(root_dir, c)
        if not os.path.isdir(charm_dir):
            continue
        log.debug("Indexing %s", c)
        try:
            charm_set = db.find({"name": c})
            for charm in charm_set:
                index_charm(indexer, charm)
            if not charm_set:
                log.info("Skipping unknown charm %s", c)
        except:
            log.exception("Unable to index %s", c)


def index_queue(db, server, indexer, index_queue="charm-index"):

    while 1:
        item = server.lpop(index_queue)
        if not item:
            return

        try:
            charm_data = json.loads(item)
            charm = db.find_one({"_id": charm_data["branch_spec"]})
            log.debug("Indexing charm %s", charm_data["branch_spec"])
            if charm:
                index_charm(indexer, charm)
            else:
                log.info(
                    "Skipping unknown charm %s", charm_data["branch_spec"])
        except:
            log.exception("Error indexing %s", item)
            continue


def main():
    logging.basicConfig(
        level=logging.WARNING,
        format="%(asctime)s: %(name)s@%(levelname)s: %(message)s")

    indexer = xappy.IndexerConnection(
        os.path.expanduser("/var/lib/charms/index"))
    setup_indexer(indexer)

    connection = pymongo.Connection()
    db = connection.juju.charms

    # Index queue
    server = redis.Redis()
    index_queue(db, server, indexer)

    # Reindex from disk repo and mongo
    #root_dir = os.path.abspath(
    #    os.path.expanduser("~/projects/charmworld/repo"))
    #index_repo(db, indexer, root_dir)
    indexer.close()
    log.info("Indexed Repo")

if __name__ == '__main__':
    main()
