import json
import logging
import os
import pymongo
import redis
import yaml


log = logging.getLogger("charm.scan")


def scan_charm(db, charm_data):
    revision_path = os.path.join(
        charm_data["branch_dir"], "revision")
    metadata_path = os.path.join(
        charm_data["branch_dir"], "metadata.yaml")
    config_path = os.path.join(
        charm_data["branch_dir"], "config.yaml")

    if not os.path.exists(metadata_path):
        log.info("Charm has no metadata: %s", charm_data["branch_spec"])
        return

    with open(metadata_path) as fh:
        try:
            metadata = yaml.load(fh.read())
        except Exception, e:
            log.info(
                "Invalid charm metadata %s: %s", charm_data["branch_spec"], e)
            return

    if os.path.exists(config_path):
        try:
            config_raw = open(config_path).read()
            config = yaml.load(config_raw)
        except:
            log.info("Invalid config metadata %s", charm_data["branch_spec"])
            return
        metadata["config"] = config
        metadata["config_raw"] = config_raw

    if os.path.exists(revision_path):
        try:
            rev_id = int(open(revision_path).read().strip())
        except:
            log.info("Invalid revision %s", charm_data["branch_spec"])
        metadata["revision"] = rev_id
    elif not "revision" in metadata:
        log.info("Invalid revision %s", charm_data["branch_spec"])
        metadata["revision"] = 0

    hook_dir = os.path.join(charm_data["branch_dir"], "hooks")
    hooks = []
    if os.path.exists(hook_dir):
        hooks = []
        for h in os.listdir(hook_dir):
            h_path = os.path.join(hook_dir, h)
            if os.path.isdir(h_path):
                continue
            if os.path.islink(h_path):
                target = os.readlink(h_path)
                target_norm = os.path.abspath(os.path.join(hook_dir, target))
                if not os.path.exists(target_norm) \
                       or not charm_data['branch_dir'] in target_norm:
                    log.debug("Invalid symlink %s %s %s" % (
                        charm_data['branch_spec'], h, target))
                    continue
            hooks.append(h)
        hooks.sort()
    metadata["hooks"] = hooks

    # Stuff into the db
    metadata.update(charm_data)
    metadata["_id"] = metadata["branch_spec"]
    item = db.find_one({"_id": metadata["_id"]})
    if item is None:
        item = metadata
    else:
        #log.debug("Updating %s", metadata["branch_spec"])
        item.update(metadata)
    item = process_charm(item)
    db.update({"_id": item["_id"]}, item, upsert=True)


def process_charm(charm):
    # Enrich charm metadata for webapp.

    # Charm url
    if charm["owner"] == "charmers":
        charm["short_url"] = "/charms/%s/%s" % (
            charm["series"], charm["name"])
    else:
        charm["short_url"] = "/~%s/%s/%s" % (charm["owner"],
                                             charm["series"],
                                             charm["name"])

    # Charm label
    if charm["owner"] == "charmers":
        charm["label"] = "%s/%s" % (charm["series"], charm["name"])
    else:
        charm["label"] = "~%s:%s/%s" % (charm["owner"],
                                        charm["series"],
                                        charm["name"])

    # Flatten the interfaces provided
    i_provides = []
    provides = charm.get("provides")
    if provides:
        for v in provides.values():
            if not isinstance(v, dict):
                continue
            i = v.get("interface")
            if not i:
                continue
            i_provides.append(i)
    charm["i_provides"] = i_provides

    # Flatten the interfaces required
    i_requires = []
    requires = charm.get("requires")
    if requires:
        for v in requires.values():
            i = v.get("interface")
            if not i:
                continue
            i_requires.append(i)
    charm["i_requires"] = i_requires
    return charm


def scan_repo(db, root_dir):
    charms = os.listdir(root_dir)
    for c in charms:
        charm_dir = os.path.join(root_dir, c)
        if not os.path.isdir(charm_dir):
            continue
        #log.info("Processing %s", c)
        try:
            scan_charm(db, c, charm_dir, repo="~charmers/charm/oneiric/%s" % c)
        except:
            log.exception("Unknown scan error")
            raise
            import pdb, traceback, sys
            traceback.print_exc()
            pdb.post_mortem(sys.exc_info()[-1])
            raise


def scan_charms(
    db, server, root_dir, scan_queue="charm-scan", index_queue="charm-index"):

    while 1:
        item = server.lpop(scan_queue)
        if not item:
            return
        try:
            charm_data = json.loads(item)
            #log.info("Processing %s", charm_data["branch_spec"])
            scan_charm(db, charm_data)
        except Exception, e:
            log.exception("scan error %s: %s", item, e)
            import pdb, traceback, sys
            traceback.print_exc()
            pdb.post_mortem(sys.exc_info()[-1])
            break
        else:
            server.lpush(index_queue, item)


def main():
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s: %(name)s@%(levelname)s: %(message)s")
    root_dir = os.path.abspath(
        os.path.expanduser("/var/lib/charms"))

    conn = pymongo.Connection()
    db = conn.juju.charms
    server = redis.Redis()
    scan_charms(db, server, root_dir)
    #scan_repo(db, root_dir)

if __name__ == '__main__':
    main()
