from datetime import datetime, timedelta
import fnmatch
import os
import logging

from webob import Response
from webhelpers import feedgenerator

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPFound
from pyramid.exceptions import NotFound

import pymongo

from models import db

import reports as report_catalog
import graph
import utils

PROJECT_NAME = u"Charm Browser"

# 15m which is the cron ingest cycle (also clears cache)
HTTP_CACHE = 950

# 3m
REPORT_CACHE = 180

log = logging.getLogger('charms')


def _found(res):
    if res is None:
        raise NotFound()
    return res

def _name_filter(n):
    if isinstance(n, list):
        n = n[0]
    idx = n.find('<')
    if idx == -1:
        return n
    return n.split("<", 1)[0].strip()


@view_config(
    route_name="home",
    http_cache=HTTP_CACHE,
    renderer="charmworld:templates/index.pt")
def home(request):
    return {}


@view_config(
    route_name="search",
    renderer="charmworld:templates/search.pt")
def search(request):
    if not "search_text" in request.params:
        url = request.route_url("home")
        return HTTPFound(location=url)
    import xappy
    searcher = xappy.SearchConnection('/var/lib/charms/index')
    query = searcher.query_parse(request.params["search_text"])
    import time
    t = time.time()
    results = query.search(0, 30)
    search_time = time.time() - t
    search_time = u"%.5f" % search_time
    return {"results": results,
            "search_time": search_time,
            "charm_total": searcher.get_doccount()}


@view_config(
    route_name="charmers",
    renderer="charmworld:templates/charmers.pt")
def charmers(request):
    # We want to collect which users have worked on which charms.
    # By worked on means they have at least one commit to it.
    # collect a list of users grouped by charm name.
    # The list of users does not correspond to the lp ids.
    results = db.charms.find(
        {}, {"name": 1, 'owner': 1, 'short_url': 1})

    owner_map = {}
    for i in results:
        if not i['owner'] in owner_map:
            owner_map[i['owner']] = 1
        else:
            owner_map[i['owner']] += 1

    owners = owner_map.items()
    owners.sort(lambda x, y: cmp(y[1], x[1]))
    owners.pop(0)
    return {"owners": owners}


def _sub_filter(d, r):
    if r.GET.get('sub'):
        d['subordinate'] = True
        return d, True
    return d, False

@view_config(
    route_name="charm-collection",
    http_cache=HTTP_CACHE,
    renderer="charmworld:templates/charm-collection.pt")
def charm_collection(request):
    query, sub_only = _sub_filter({"owner": "charmers"}, request)
    charms = _found(db.charms.find(
        query,
        {"name": 1, "owner": 1, "series": 1, "summary": 1, "short_url": 1},
        sort=[("name", pymongo.ASCENDING)]))
    return {"charms": list(charms), 'sub_only': sub_only}


@view_config(
    route_name="personal-collection",
    http_cache=HTTP_CACHE,
    renderer="charmworld:templates/charm-collection.pt")
def personal_collection(request):
    query, sub_only = _sub_filter({"owner": request.matchdict["owner"]}, request)
    charms = _found(db.charms.find(
        query,
        {"name": 1, "owner": 1, "series": 1, "summary": 1, "short_url": 1},
        sort=[("name", pymongo.ASCENDING)]))
    return {"charms": list(charms), "sub_only": sub_only}


@view_config(
    route_name="series-collection",
    http_cache=HTTP_CACHE,
    renderer="charmworld:templates/charm-collection.pt")
def series_collection(request):
    query, sub_only = _sub_filter(
        {"owner": "charmers", "series": request.matchdict["series"]}, request)
    charms = _found(db.charms.find(
        query,
        {"name": 1, "owner": 1, "series": 1, "summary": 1, "short_url": 1,},
        sort=[("name", pymongo.ASCENDING)]))
    return {"charms": list(charms), 'sub_only': sub_only}


@view_config(
    route_name="personal-series",
    http_cache=HTTP_CACHE,
    renderer="charmworld:templates/charm-collection.pt")
def personal_series_collection(request):
    query, sub_only = _sub_filter(
        {"owner": request.matchdict["owner"], "series": request.matchdict["series"]},
        request)

    charms = _found(db.charms.find(
        query,
        {"name": 1, "owner": 1, "series": 1, "summary": 1, "short_url": 1},
        sort=[("name", pymongo.ASCENDING)]))
    return {"charms": list(charms), 'sub_only': sub_only}


@view_config(
    route_name="personal-charm",
    http_cache=HTTP_CACHE,
    renderer="charmworld:templates/charm.pt")
def personal_charm(request):
    charm = _found(db.charms.find_one(
        {"name": request.matchdict['charm'],
         "owner": request.matchdict['owner'],
         "series": request.matchdict['series']}))
    return _charm_view(charm)


@view_config(
    route_name="charm",
    http_cache=HTTP_CACHE,
    renderer="charmworld:templates/charm.pt")
def distro_charm(request):
    charm = _found(db.charms.find_one(
        {"name": request.matchdict["charm"],
         "owner": "charmers",
         "series": request.matchdict["series"]}))
    return _charm_view(charm)


@view_config(route_name="personal-charm-changes", http_cache=HTTP_CACHE)
def personal_charm_changes(request):
    charm = _found(db.charms.find_one(
        {"name": request.matchdict['charm'],
         "owner": request.matchdict['owner'],
         "series": request.matchdict['series']}))
    feed = _charm_changes_feed(charm, request)
    return Response(feed.writeString("utf8"),
                    content_type=u"application/atom+xml")


@view_config(route_name="charm-changes", http_cache=HTTP_CACHE)
def charm_changes(request):
    charm = _found(db.charms.find_one(
        {"name": request.matchdict['charm'],
         "owner": u"charmers",
         "series": request.matchdict['series']}))
    feed = _charm_changes_feed(charm, request)
    return Response(feed.writeString("utf8"),
                    content_type=u"application/atom+xml")


def _charm_changes_feed(charm, request):
    feed = feedgenerator.Atom1Feed(
        title=u"Charm browser changes",
        description="Recently changed charms",
        link=request.url,
        language=u"en"
        )

    for change in charm['changes']:
        committer = _name_filter(change['committer'])

        diff_link = u"https://bazaar.launchpad.net/%s/revision/%s" % (
            charm["branch_spec"], change["revno"])

        feed.add_item(
            title=charm['name'],
            link=diff_link,
            description=change['message'].strip().replace("\n", "<br/>"),
            author_name=committer,
            pub_date=feedgenerator.rfc3339_date(
                datetime.fromtimestamp(change['created'])))
    return feed


def format_change(change, charm=None):
    """Format a bzr commit."""
    parts = []
    date = datetime.fromtimestamp(change["created"])
    parts.append(unicode(date.strftime("%Y/%m/%d")))
    parts.append(_name_filter(change['committer']))
    parts.append(change["message"][:100])
    if charm is not None:
        parts.append(
            u'(<a href="'
            'https://bazaar.launchpad.net/%s/revision/%s">revno %s</a>)' % (
                charm["branch_spec"], change["revno"], change["revno"]))
    return u" ".join(parts)


def format_proof(charm):
    proof = charm['proof']

    messages = []
    for e in proof.get('c', []):
        messages.append('crit: %s' % e)
    for e in proof.get('e', []):
        messages.append("err: %s" % e)
    for e in proof.get('w', []):
        messages.append('warn: %s' % e)
    return messages


def _charm_view(charm):
    # Compute display values for requires, provides, and config.
    charm["d_requires"] = d_requires = []

    # Defensive parsing, i know.. ingest time would have been nice to weed these out,
    # but we want to show/report things that are broken and flag them appropriately.
    requires = charm.get('requires', {})
    if requires is None:
        log.warning("charm with invalid requires %s", charm['branch_spec'])
        requires = {}
    for k, v in requires.items():
        d = {"name": k, "interface": v["interface"]}
        d_requires.append(d)

    charm["d_provides"] = d_provides = []
    for k, v in charm.get("provides", {}).items():

        if isinstance(v, basestring):
            interface = v
        else:
            interface = v['interface']
        d = {"name": k, "interface": interface}
        d_provides.append(d)

    charm["d_config"] = d_config = []

    if charm.get('config', {'options': {}}).get('options'):
        for k, v in charm.get("config", {"options": {}})["options"].items():
            d = {"name": k,
                 "default": v.get('default', u''),
                 'type': v.get('type', u''),
                 'description': v.get('description', u'')}
            d_config.append(d)

    # Find the same package in other archives.
    others = list(db.charms.find(
        {"name": charm["name"], "series": charm["series"],
         "branch_spec": {"$ne": charm["branch_spec"]}}))

    readme = _charm_readme(charm)
    charm['maintainer'] =  _name_filter(charm.get('maintainer', ''))
    return {"project": PROJECT_NAME, "charm": charm, "name": charm["name"],
            "others": others, 'format_change': format_change, "readme": readme,
            'format_proof': format_proof}



def _charm_readme(charm):
    # Find a readme
    dirents = os.listdir(charm['branch_dir'])
    dirents = dict([(k.lower(), v) for k, v in zip(dirents, dirents)])
    matches = fnmatch.filter(dirents.keys(), "readme*")
    if not matches:
        return None
    matches.sort(lambda x, y: cmp(len(x), len(y)))
    readme_path = os.path.join(charm['branch_dir'], dirents[matches[0]])

    with open(readme_path) as fh:
        return unicode(fh.read(), "utf8")


@view_config(
    route_name="hook",
    http_cache=HTTP_CACHE,
    renderer="charmworld:templates/hook.pt")
def hook(request):
    charm = _found(db.charms.find_one({
        "owner": "charmers",
        "series": request.matchdict["series"],
        "name": request.matchdict["charm"]}))
    return _hook_view(charm, request.matchdict["hook"])


@view_config(
    route_name="personal-hook",
    http_cache=HTTP_CACHE,
    renderer="charmworld:templates/hook.pt")
def personal_hook(request):
    charm = _found(db.charms.find_one({
        "owner": request.matchdict["owner"],
        "series": request.matchdict["series"],
        "name": request.matchdict["charm"]}))
    return _hook_view(charm, request.matchdict["hook"])


def _hook_view(charm, hook):
    hook_path = os.path.join(charm["branch_dir"], "hooks", hook)
    data =  {"project": PROJECT_NAME,
             "hook_raw": "Hook not found!",
             "charm": charm,
             "hook": hook,
             "onload": "prettyPrint()"}

    if not os.path.exists(hook_path):
        log.warning("Hook not found %s %s %s" % (
            charm['branch_spec'],
            hook,
            hook_path))
        return data

    with open(hook_path) as fh:
        hook_raw = fh.read()
        data['hook_raw'] = hook_raw
    return data


@view_config(
    route_name="config",
    http_cache=HTTP_CACHE,
    renderer="charmworld:templates/config.pt")
def config(request):
    charm = _found(db.charms.find_one({
        "owner": "charmers",
        "series": request.matchdict["series"],
        "name": request.matchdict["charm"]}))
    return _config_view(charm)


@view_config(
    route_name="personal-config",
    http_cache=HTTP_CACHE,
    renderer="charmworld:templates/config.pt")
def personal_config(request):
    charm = _found(db.charms.find_one({
        "owner": request.matchdict["owner"],
        "series": request.matchdict["series"],
        "name": request.matchdict["charm"]}))
    return _config_view(charm)


def _config_view(charm):
    return {"charm": charm, "onload": "prettyPrint()"}


@view_config(
    route_name="tools-review",
    http_cache=REPORT_CACHE,
    renderer="charmworld:templates/review.pt")
def charm_review(request):
    # This view's data in a separate cron job than ingest
    queue = list(
        db.review_queue.find(
            sort=[('date_created', pymongo.ASCENDING)]))
    return {'queue': queue, 'age_format': utils.pretty_timedelta}

@view_config(
    route_name="tools-core-review",
    http_cache=REPORT_CACHE,
    renderer="charmworld:templates/review.pt")
def core_review(request):
    # This view's data in a separate cron job than ingest
    queue = list(
        db.core_review_queue.find(
            sort=[('date_created', pymongo.ASCENDING)]))
    return {'queue': queue, 'age_format': utils.pretty_timedelta}

@view_config(
    route_name="tools-store-missing",
    http_cache=REPORT_CACHE,
    renderer="charmworld:templates/store-missing.pt")
def charm_store_missing(request):
    queue = list(
        db.charms.find(
            {"store_data.errors": {"$exists": True}}
            ))

    return {'queue': queue, 'age_format': utils.pretty_timedelta}

@view_config(
    route_name="tools-maintainer-missing",
    http_cache=REPORT_CACHE,
    renderer="charmworld:templates/charm-collection.pt")
def charm_maintainer_missing(request):
    charms = list(
        db.charms.find(
            _sub_filter({"maintainer": {"$exists": False},
                         "owner": "charmers"},
                        request),
            sort=[("series", pymongo.DESCENDING), ("name", pymongo.ASCENDING)]
            ))

    return {'charms': charms, 'age_format': utils.pretty_timedelta}

@view_config(
    route_name="tools-proof-errors",
    http_cache=REPORT_CACHE,
    renderer="charmworld:templates/charm-proof-errors.pt")
def proof_errors(request):
    query, sub_only = _sub_filter(
        {"proof.e": {"$exists": True}, "owner": "charmers"}, request)

    charms = list(
        db.charms.find(
            query,
            sort=[("series", pymongo.DESCENDING), ("name", pymongo.ASCENDING)]
            ))

    return {'charms': charms, 'sub_only': sub_only, 'format_proof': format_proof}


@view_config(
    route_name="interface-collection-map",
    http_cache=HTTP_CACHE,
    renderer="charmworld:templates/interface-collection.pt")
@view_config(
    route_name="interface-collection",
    http_cache=HTTP_CACHE,
    renderer="charmworld:templates/interface-collection-redux.pt")
def interface_collection(request):
    return {"project": PROJECT_NAME, "interfaces": _interfaces()}


def _interfaces():
    interfaces = {}

    # This is all static between loads
    results = db.charms.find(
        {}, {"name": 1, "i_provides": 1, "i_requires": 1,
             'label': 1, 'short_url': 1, 'owner': 1})

    for i in results:
        for p in i["i_provides"]:
            if not p in interfaces:
                interfaces[p] = i_info = {
                    "name": p, "provided_by": [], "required_by": []}
            else:
                i_info = interfaces[p]
            i_info["provided_by"].append(i)

        for d in i["i_requires"]:
            if not d in interfaces:
                interfaces[d] = i_info = {
                    "name": d, "provided_by": [], "required_by": []}
            else:
                i_info = interfaces[d]
            i_info["required_by"].append(i)

    # sort them
    interfaces = interfaces.values()
    interfaces.sort(lambda x, y: cmp(x["name"], y["name"]))

    for i in interfaces:
        i["provided_by"] = _unique_sort(i["provided_by"])
        i["required_by"] = _unique_sort(i["required_by"])

    return interfaces


def _unique_sort(charms):
    names = []
    results = []
    for c in charms:
        if c["name"] in names:
            if c['owner'] != 'charmers':
                continue
            else:
                idx = names.index(c['name'])
                results.pop(idx)
                names.pop(idx)
        results.append(c)
        names.append(c["name"])
    results.sort(lambda x, y: cmp(x["name"], y["name"]))
    return results


@view_config(
    route_name="interface",
    renderer="charmworld:templates/interface.pt",
    http_cache=HTTP_CACHE)
def interface(request):
    iface = request.matchdict["interface"]
    return {
        "interface": iface,
        "providers": list(db.charms.find({"i_provides": iface},
                                    sort=[("name", pymongo.ASCENDING)])),
        "requirers": list(db.charms.find({"i_requires": iface},
                                    sort=[("name", pymongo.ASCENDING)]))}


def _recently_changed():
    charms = db.charms.find({}, {"last_change": 1,
                                 "label": 1,
                                 "name": 1,
                                 "short_url": 1,
                                 "summary": 1,
                                 "branch_spec": 1},
                           sort=[("last_change.created", pymongo.DESCENDING)],
                           limit=50)
    return {"charms": charms, 'format_change': format_change}


@view_config(
    route_name="recently-changed",
    renderer="charmworld:templates/recent.pt",
    http_cache=int(HTTP_CACHE / 2))
def recently_changed(request):
    return _recently_changed()


@view_config(route_name="recently-changed-feed", http_cache=HTTP_CACHE)
def recently_changed_feed(request):
    feed = feedgenerator.Atom1Feed(
        title=u"Charm browser changes",
        description="Recently changed charms",
        link=request.url,
        language=u"en"
        )

    for charm in  _recently_changed()['charms']:
        change = charm['last_change']
        committer = _name_filter(change['committer'])

        diff_link = u"https://bazaar.launchpad.net/%s/revision/%s" % (
            charm["branch_spec"], change["revno"])

        feed.add_item(
            title=charm['name'],
            link=diff_link,
            description=change['message'].strip().replace("\n", "<br/>"),
            author_name=committer,
            pub_date=feedgenerator.rfc3339_date(
                datetime.fromtimestamp(change['created'])))

    return Response(feed.writeString("utf8"),
                    content_type=u"application/atom+xml")


# Reports
@view_config(
    route_name="reports", renderer="charmworld:templates/reports.pt")
def reports(request):
    return {}


@view_config(
    route_name="json-charm-change",
    renderer="json",
    http_cache=HTTP_CACHE)
def json_charm_changes(request):
    return report_catalog.series_charm_changes(step=timedelta(7)).records()


@view_config(
    route_name="json-charm-growth",
    renderer="json",
    http_cache=HTTP_CACHE)
def json_charm_growth(request):
    return report_catalog.series_charm_growth().records()


@view_config(
    route_name="json-contributor-growth",
    renderer="json",
    http_cache=HTTP_CACHE)
def json_contributor_growth(request):
    return report_catalog.series_charm_authors().records()


@view_config(
    route_name="json-interface-graph",
    #renderer="json",
    http_cache=HTTP_CACHE)
def json_interface_graph(request):
    import json
    return Response(
        json.dumps(graph.graph_by_charms(db.charms.find(), _interfaces()),
                   indent=2),
        content_type=u"application/json")


@view_config(
    route_name="charm-graph",
    renderer="charmworld:templates/world-graph.pt",
    )
def charm_graph(request):
    return {}
